/* GTK - Spell widget
 * Copyright (C) 1997 Paolo Molaro
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <gtk/gtk.h>
#include "gtkspell.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>

enum {
	FOUND_WORD,
	REPLACE_WORD,
	LAST_SIGNAL
};

typedef void (*GtkSpellFunction) (GtkObject* obj, gpointer arg1, gpointer data);

static void gtk_spell_class_init(GtkSpellClass* klass);
static void gtk_spell_init(GtkSpell* spell);

static void gtk_spell_browse_handler(GtkButton * button, GtkEntry* entry);
static void gtk_spell_enable_tooltips(GtkToggleButton* b, GtkSpell* spell);
static void gtk_spell_marshaller (GtkObject* obj, GtkSignalFunc func, gpointer func_data, GtkArg* args);
static void gtk_spell_start_proc(GtkSpell* spell);
static void gtk_spell_send_string(GtkSpell* spell, gchar* s);
static gchar* gtk_spell_read_string(GtkSpell* spell);
static void gtk_spell_fill_info(GtkSpellInfo* sp, gchar* src);
static void gtk_spell_found_word(GtkSpell* spell, GtkSpellInfo* sp);
static void gtk_spell_button_handler(GtkButton* button, GtkSpell* spell);
static void gtk_spell_list_handler(GtkButton* button, GtkListItem* item, GtkSpell* spell);

static gint gtk_spell_signals[LAST_SIGNAL] = { 0 };

/*static GtkVBoxClass* parent_class = 0;*/

static void
gtk_spell_marshaller (GtkObject* obj, GtkSignalFunc func, gpointer func_data, GtkArg* args) {
	GtkSpellFunction rfunc;

	rfunc = (GtkSpellFunction) func;
	(*rfunc) (obj, args[0].d.p, func_data);
}

static void
gtk_spell_browse_handler(GtkButton * button, GtkEntry* entry) {

}

static void 
gtk_spell_enable_tooltips(GtkToggleButton* b, GtkSpell *spell) {

	if ( b->active )
		gtk_tooltips_enable(spell->tooltips);
	else
		gtk_tooltips_disable(spell->tooltips);
}

static void 
gtk_spell_button_handler(GtkButton* button, GtkSpell* spell) {
	gchar* word = gtk_entry_get_text(GTK_ENTRY(spell->word));
	if ( ! word || !*word )
		return;
	if ( button == GTK_BUTTON(spell->b_accept) )
		gtk_spell_accept(spell, word);
	else if ( button == GTK_BUTTON(spell->b_replace) ) {
		if ( spell->spellinfo ) {
			GtkSpellInfo * sp = (GtkSpellInfo*)spell->spellinfo->data;
			sp->replacement = g_strdup(gtk_entry_get_text(GTK_ENTRY(spell->word)));
			gtk_signal_emit(GTK_OBJECT(spell), 
				gtk_spell_signals[REPLACE_WORD], sp);
		}
	} else if ( button != GTK_BUTTON(spell->b_skip) )
		gtk_spell_insert(spell, word, button == GTK_BUTTON(spell->b_insertl));
	/* button "skip" jumps here */
	gtk_entry_set_text(GTK_ENTRY(spell->context), "");
	gtk_entry_set_text(GTK_ENTRY(spell->word), "");
	gtk_list_remove_items(GTK_LIST(spell->list), (GTK_LIST(spell->list))->children);
	if ( spell->spellinfo ) {
		GtkSpellInfo * sp = (GtkSpellInfo*)spell->spellinfo->data;
		gtk_spell_fill_info(sp, NULL);
		spell->spellinfo = g_slist_remove(spell->spellinfo, sp);
	}
	if ( spell->spellinfo ) {
		gtk_signal_emit(GTK_OBJECT(spell),
			gtk_spell_signals[FOUND_WORD], spell->spellinfo->data);
	}
}

static void 
gtk_spell_list_handler(GtkButton* button, GtkListItem* item, GtkSpell* spell) {
	gchar* word;
	gtk_label_get( GTK_LABEL( GTK_BIN(item)->child ), &word);
	gtk_entry_set_text(GTK_ENTRY(spell->word), word);
}

static void 
gtk_spell_start_proc(GtkSpell* spell) {
	int fd1[2];
	int fd2[2];
	char* cmd = "ispell"; /* make ./configure find this */
	char* args[16] = { 
		"gtkspellchecker",
		"-a",
		0
	};
	char* arg;
	int argpos = 2;
	if ( (arg=gtk_entry_get_text(GTK_ENTRY(spell->dict))) && *arg ) {
		args[argpos++] = "-d";
		args[argpos++] = arg;
	}
	if ( (arg=gtk_entry_get_text(GTK_ENTRY(spell->pdict))) && *arg ) {
		args[argpos++] = "-p";
		args[argpos++] = arg;
	}
	if ( (arg=gtk_entry_get_text(GTK_ENTRY(spell->wchars))) && *arg ) {
		args[argpos++] = "-w";
		args[argpos++] = arg;
	}
	if ( (arg=gtk_entry_get_text(GTK_ENTRY(spell->wlen))) && *arg ) {
		args[argpos++] = "-W";
		args[argpos++] = arg;
	}
	if ( (GTK_TOGGLE_BUTTON(spell->sort))->active )
		args[argpos++] = "-S";
	if ( (GTK_TOGGLE_BUTTON(spell->runtog))->active )
		args[argpos++] = "-C";

	if ( pipe(fd1) || pipe(fd2) ) {
		perror("gtkspell: pipe()");
		return;
	}

	/* install signal handlers */
	spell->spell_pid = fork();
	if ( spell->spell_pid < 0 ) {
		perror("gtkspell: fork()");
		return;
	} else if ( spell->spell_pid == 0 ) {
		close(fd1[0]);
		close(fd2[1]);
		if ( dup2(fd2[0], 0) < 0 ) {
			perror("gtkspell: dup2()");
			_exit(1);
		}
		if ( dup2(fd1[1], 1) < 0 ) {
			perror("gtkspell: dup2()");
			_exit(1);
		}
		execvp(cmd, args);
		perror("gtkspell: exec()");
		_exit(1);
	} else if ( spell->spell_pid > 0 ) {
		char* spellid;
		close(fd2[0]);
		close(fd1[1]);
		spell->rispell = fdopen(fd1[0], "r");
		if ( !spell->rispell )
			perror("gtkspell: r-fdopen()");
		spell->wispell = fdopen(fd2[1], "w");
		if ( !spell->wispell )
			perror("gtkspell: w-fdopen()");
		spellid = gtk_spell_read_string(spell);
		if ( !spellid || !strstr(spellid, "spell") ) {
			fprintf(stderr, "What spell is this?\n"); /* cut the check for "spell"? */
		}
		g_free(spellid);
		gtk_spell_send_string(spell, "!\n");
	}

}

static void 
gtk_spell_send_string(GtkSpell* spell, gchar* s) {
	if ( spell->spell_pid <= 0 /* || kill(spell->spell_pid, SIGCONT) */)
		gtk_spell_start_proc(spell);
	fprintf(spell->wispell, "%s", s);
	fflush(spell->wispell);
#ifdef TEST
	fprintf(stderr, "WRITE: %s", s);
#endif
}

static gchar* 
gtk_spell_read_string(GtkSpell* spell) {
	gchar* buf;
	guint buflen=1024;
	if ( spell->spell_pid <= 0 || ! spell->rispell )
		return NULL;
	buf = alloca(buflen);
	/* select or gdk_input_add? */
	if ( !fgets(buf, buflen, spell->rispell) )
		return NULL;
	/* discard longer lines? */
#ifdef TEST
	fprintf(stderr, "READ: %s", buf);
#endif
	return g_strdup(buf);
}

static void
build_page_spell(GtkSpell* spell) {
	GtkWidget* hbox;
	GtkWidget* button;
	GtkWidget* label;
	GtkWidget* list;
	GtkWidget* scwindow;
	GtkWidget* entry;
	GtkWidget* vbox;
	GtkWidget* page_spell=spell->page_spell;

	hbox = gtk_hbox_new(0, 5);
	gtk_widget_show(hbox);
	label = gtk_label_new("Context:");
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(page_spell), hbox);
	spell->context = entry;

	hbox = gtk_hbox_new(0, 5);
	gtk_widget_show(hbox);
	label = gtk_label_new("Word:");
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(page_spell), hbox);
	spell->word = entry;

	hbox = gtk_hbox_new(0, 5);
	gtk_widget_show(hbox);
	
	vbox = gtk_vbox_new(0, 5);
	gtk_widget_show(vbox);
	label = gtk_label_new("Alternatives:");
	gtk_widget_show(label);
	scwindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_show(scwindow);
	list = gtk_list_new();
	gtk_widget_show(list);
	gtk_signal_connect(GTK_OBJECT(list), "select_child",
		(GtkSignalFunc)gtk_spell_list_handler, spell);
	gtk_container_add(GTK_CONTAINER(scwindow), list);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), scwindow, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
	spell->list = list;

	vbox = gtk_vbox_new(0, 5);
	gtk_widget_show(vbox);
	button = gtk_button_new_with_label("Accept");
	gtk_widget_show(button);
	spell->b_accept = button;
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_spell_button_handler, spell);
	gtk_tooltips_set_tips(spell->tooltips, button, "Accept the word for\nthis session only");
	gtk_container_add(GTK_CONTAINER(vbox), button);
	button = gtk_button_new_with_label("Skip");
	gtk_widget_show(button);
	spell->b_skip = button;
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_spell_button_handler, spell);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	button = gtk_button_new_with_label("Replace");
	gtk_widget_show(button);
	spell->b_replace = button;
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_spell_button_handler, spell);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	button = gtk_button_new_with_label("Insert");
	gtk_widget_show(button);
	spell->b_insertl = button;
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_spell_button_handler, spell);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	button = gtk_button_new_with_label("Insert with case");
	gtk_widget_show(button);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	spell->b_insert = button;
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_spell_button_handler, spell);
	gtk_box_pack_end(GTK_BOX(hbox), vbox, TRUE, FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(page_spell), hbox);
}

static void
build_page_config(GtkSpell* spell) {
	GtkWidget* label;
	GtkWidget* hbox;
	GtkWidget* entry;
	GtkWidget* button;
	GtkWidget* page_config=spell->page_config;

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	label = gtk_label_new("Main dictionary:");
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	gtk_entry_set_text(GTK_ENTRY(entry), "default");
	spell->dict = entry;
	/* browse button ... */
	button = gtk_button_new_with_label("Browse");
	gtk_widget_show(button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", 
		(GtkSignalFunc) gtk_spell_browse_handler, entry);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	label = gtk_label_new("Personal dictionary:");
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	spell->pdict = entry;
	/* browse button ... */
	button = gtk_button_new_with_label("Browse");
	gtk_widget_show(button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", 
		(GtkSignalFunc) gtk_spell_browse_handler, entry);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	label = gtk_label_new("Valid chars in words:");
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	spell->wchars = entry;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	label = gtk_label_new("Check words longer than:");
	gtk_widget_show(label);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	spell->wlen = entry;
	gtk_entry_set_text(GTK_ENTRY(entry), "1");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	button = gtk_check_button_new_with_label("Sort by correctness");
	gtk_widget_show(button);
	spell->sort = button;
	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	button = gtk_check_button_new_with_label("Accept run-together words");
	gtk_widget_show(button);
	spell->runtog = button;
	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	/* include guesses */

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	button = gtk_check_button_new_with_label("Enable tooltips");
	gtk_widget_show(button);
	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);
	gtk_signal_connect(GTK_OBJECT(button), "toggled",
		(GtkSignalFunc)gtk_spell_enable_tooltips, spell);

	hbox = gtk_hseparator_new();
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	button = gtk_button_new_with_label("Kill ispell process");
	gtk_widget_show(button);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, FALSE, 0);
	button = gtk_button_new_with_label("Apply options");
	gtk_widget_show(button);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, FALSE, 0);
	button = gtk_button_new_with_label("Save options");
	gtk_widget_show(button);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(page_config), hbox);

}

static GtkWidget*
build_widget(GtkSpell* spell) {
	GtkWidget* notebook;
	GtkWidget* label;

	notebook = gtk_notebook_new();
	/*gtk_container_border_width(GTK_CONTAINER(notebook), 10);*/

	spell->page_spell = gtk_vbox_new(FALSE, 5);
	gtk_container_border_width(GTK_CONTAINER(spell->page_spell), 10);
	gtk_widget_show(spell->page_spell);
	build_page_spell(spell);
	label = gtk_label_new("Spell");
	gtk_widget_show(label);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), spell->page_spell, label);
	
	spell->page_config = gtk_vbox_new(FALSE, 5);
	gtk_widget_show(spell->page_config);
	build_page_config(spell);
	label = gtk_label_new("Configure");
	gtk_widget_show(label);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), spell->page_config, label);
	
	gtk_widget_show(notebook);
	
	return notebook;
}

static void
gtk_spell_destroy (GtkObject* s) {
	GtkSpell* spell = (GtkSpell*)s;
	if ( spell->tooltips )
		gtk_tooltips_destroy(spell->tooltips);
	if ( spell->rispell )
		fclose(spell->rispell);
	if ( spell->wispell )
		fclose(spell->wispell);
	if ( spell->spell_pid > 0 )
		waitpid(spell->spell_pid, NULL, WNOHANG | WUNTRACED);
	if ( spell->spellinfo ) {
		g_slist_foreach(spell->spellinfo, (GFunc)gtk_spell_fill_info, NULL);
		g_slist_free(spell->spellinfo);
	}
}

guint
gtk_spell_get_type () {
	static guint gtk_spell_type = 0;

	if ( !gtk_spell_type ) {
		GtkTypeInfo spell_info = {
			"GtkSpell",
			sizeof(GtkSpell),
			sizeof(GtkSpellClass),
			(GtkClassInitFunc) gtk_spell_class_init,
			(GtkObjectInitFunc) gtk_spell_init,
		};
		gtk_spell_type = gtk_type_unique(gtk_vbox_get_type(), &spell_info);
	}
	return gtk_spell_type;
}

static void
gtk_spell_class_init(GtkSpellClass* klass) {
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) klass;

	gtk_spell_signals[FOUND_WORD] = gtk_signal_new ("found_word",
		GTK_RUN_FIRST,
		object_class->type,
		GTK_SIGNAL_OFFSET(GtkSpellClass, found_word),
		gtk_spell_marshaller, GTK_ARG_NONE, 1, GTK_ARG_POINTER);

	gtk_spell_signals[REPLACE_WORD] = gtk_signal_new ("replace_word",
		GTK_RUN_FIRST,
		object_class->type,
		GTK_SIGNAL_OFFSET(GtkSpellClass, replace_word),
		gtk_spell_marshaller, GTK_ARG_NONE, 1, GTK_ARG_POINTER);

	gtk_object_class_add_signals(object_class, gtk_spell_signals, LAST_SIGNAL);

	object_class->destroy = gtk_spell_destroy;

	klass->found_word = gtk_spell_found_word;
	klass->replace_word = NULL;
}

static void
gtk_spell_init(GtkSpell* spell) {
	GtkWidget * noteb;

	spell->spell_pid = -1;
	spell->rispell = spell->wispell = NULL;
	spell->tooltips = gtk_tooltips_new();
	spell->spellinfo = NULL;

	gtk_tooltips_disable(spell->tooltips);

	noteb = build_widget(spell);
	gtk_container_add(GTK_CONTAINER(spell), noteb);
}

GtkWidget*
gtk_spell_new() {
	return GTK_WIDGET(gtk_type_new(gtk_spell_get_type()));
}

/* used also for freeing sp->* stuff with src == NULL */
static void
gtk_spell_fill_info(GtkSpellInfo* sp, gchar* src) {
	gchar* p;
	gchar * reject = " ,\t\n\r\v:";
	gint count;

	sp->offset = 0;
	if ( sp->word ) {
		g_free(sp->word);
		sp->word = NULL;
	}
	if ( sp->replacement ) {
		g_free(sp->replacement);
		sp->replacement = NULL;
	}
	if ( sp->words ) {
		g_slist_foreach(sp->words, (GFunc)g_free, NULL);
		g_slist_free(sp->words);
		sp->words = NULL;
	}

	if (!src ) return;
	p = src;
	while( *p && isspace(*p) ) ++p;
	if ( *p != '&' && *p != '?' && *p != '#' ) {
		g_warning("Ispell output corrupt!\n");
		/* FIXME: se arriva qui SIGSEV */
		return;
	}
	++p;
	p = strtok(p, reject);
	if ( !p ) return;
	sp->word = g_strdup(p);
	if ( *src == '#' ) {
		sp->offset = atoi(p);
		if ( sp->offset > 0 ) --sp->offset;
	} else {
		p = strtok(NULL, reject);
		if ( p ) {
			count = atoi(p);
		} else return;
		p = strtok(NULL, reject);
		if ( p ) {
			sp->offset = atoi(p);
			if ( sp->offset > 0 ) --sp->offset;
		} else return;
		while ( (p=strtok(NULL, reject)) )
			sp->words = g_slist_append(sp->words, g_strdup(p));
	}
}

/* GtkSpell deve avere due segnali: uno indica quando sono state
   trovate le alternative alla parola sbagliata e uno quando viene premuto
   il bottone replace.
   Nel primo caso esiste una funzione di GtkSpellClass->func che
   gestisce il segnale e va a modificare la spell->list, spell->word etc.
   Perche' funzioni spell->sinfo deve diventare una lista e, mano a mano
   che le parole vengono sostituite, skippate etc. si passa a quella
   successiva...
   FIXME: C'e' il problema di quando la stessa parola sbagliata viene
   trovata piu' volte nella stessa riga e alla prima viene accettata,
   ma si ripresenta perche' l'output da ispell si riferiva a prima....
   */
void
gtk_spell_check(GtkSpell* spell, gchar* str) {
	GtkSpellInfo *sp;
	char* buf;
	char* result;

	if ( !str || !*str )
		return;
	buf = alloca(strlen(str)+2);
	/* check "\n" */
	sprintf(buf, "^%s\n", str);
	result = strchr(buf, '\n');
	++result;
	*result = 0;
	gtk_spell_send_string(spell, buf);
	while ( (result=gtk_spell_read_string(spell)) && strcmp(result, "\n") ) {
		sp = g_new(GtkSpellInfo, 1);
		sp->original = str;
		sp->word = sp->replacement = NULL;
		sp->words = NULL;
		gtk_spell_fill_info(sp, result);
		spell->spellinfo = g_slist_append(spell->spellinfo, sp);
		g_free(result);
	}
	g_free(result);
	if ( spell->spellinfo ) {
		gtk_signal_emit(GTK_OBJECT(spell),
			gtk_spell_signals[FOUND_WORD], spell->spellinfo->data);
	}
}

static void 
gtk_spell_found_word(GtkSpell* spell, GtkSpellInfo* sp) {
	GtkWidget* item;
	GSList* list;
	gtk_list_remove_items(GTK_LIST(spell->list), (GTK_LIST(spell->list))->children);
	gtk_entry_set_text(GTK_ENTRY(spell->word), sp->word);
	gtk_entry_set_text(GTK_ENTRY(spell->context), sp->original);
	gtk_entry_set_position(GTK_ENTRY(spell->context), sp->offset);
	list = sp->words;
	while (list) {
		item = gtk_list_item_new_with_label(list->data);
		gtk_widget_show(item);
		gtk_container_add(GTK_CONTAINER(spell->list), item);
		list = list->next;
	}
}

/* FIXME: in accept and insert warn if word contains any
	character other then [-a-zZ-Z0-9] and return.
*/
void
gtk_spell_accept(GtkSpell *spell, gchar* word) {
	gchar* buf = alloca(strlen(word)+2);
	sprintf(buf, "@%s\n", word);
	/* mettere in awords */
	gtk_spell_send_string(spell, buf);
}

void
gtk_spell_insert(GtkSpell* spell, gchar* word, gint lowercase) {
	gchar* w = alloca(strlen(word)+1);
	gchar* buf = alloca(strlen(word)+2);
	strcpy(w, word);
	if ( lowercase ) {
		char* p=w;
		while (*p) *p = tolower(*p);
	}
	/*  if ispell dies in a bad way it dosen't save the dictionary.
		save it at every update? (append "#\n" to the string below) */
	sprintf(buf, "*%s\n", w);
	gtk_spell_send_string(spell, buf);
}

#ifdef TEST

static void 
test_quit() {
	gtk_exit(0);
}

static void
test_check(GtkButton* b, GtkEntry* entry) {
	GtkSpell* spell;
	spell = gtk_object_get_user_data(GTK_OBJECT(entry));
	if (!spell) return;
	gtk_spell_check(spell, gtk_entry_get_text(entry));
}

static void
test_replace(GtkSpell* spell, GtkSpellInfo* sinfo, GtkEntry* entry) {
	char *buf;
	int diff;
	if (!sinfo->replacement)
		return;
	diff = strlen(sinfo->replacement) - strlen(sinfo->word);
	buf = alloca(strlen(sinfo->original)+diff+1);
	memcpy(buf, sinfo->original, sinfo->offset);
	strcpy(buf+sinfo->offset, sinfo->replacement);
	strcat(buf, sinfo->original+sinfo->offset+strlen(sinfo->word));
	gtk_entry_set_text(entry, buf);
}

int main(int argc, char* argv[]) {
	GtkWidget * mainw;
	GtkWidget * vbox;
	GtkWidget * entry;
	GtkWidget * button;
	GtkWidget* spell;
	
	gtk_init(&argc, &argv);
	gtk_rc_parse("gtkrc");

	mainw = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	vbox = gtk_vbox_new(FALSE, 10);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(mainw), vbox);
	spell = gtk_spell_new();
	gtk_widget_show(spell);
	gtk_container_add(GTK_CONTAINER(vbox), spell);
	entry = gtk_entry_new();
	gtk_widget_show(entry);
	gtk_container_add(GTK_CONTAINER(vbox), entry);
	gtk_object_set_user_data(GTK_OBJECT(entry), spell);
	button = gtk_button_new_with_label("Spell-check text");
	gtk_widget_show(button);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", 
		(GtkSignalFunc)test_check, entry);
	button = gtk_button_new_with_label("Quit");
	gtk_widget_show(button);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	gtk_widget_show(mainw);

	gtk_signal_connect(GTK_OBJECT(mainw), "destroy", 
		(GtkSignalFunc)test_quit, NULL);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", 
		(GtkSignalFunc)test_quit, NULL);
	gtk_signal_connect(GTK_OBJECT(spell), "replace_word", 
		(GtkSignalFunc)test_replace, entry);

	gtk_main();
	return 0;
}

#endif
